Public Delegate Function MyFxDelegate(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double) As Double
Public Delegate Function SayFxDelegate() As String

Public Class COptimBFGS1
  Dim m_MyFx As MyFxDelegate

  Protected Function MyFxEx(ByVal nNumVars As Integer, _
              ByRef fX() As Double, ByRef fParam() As Double, _
              ByRef fDeltaX() As Double, ByVal fLambda As Double) As Double
    Dim I As Integer
    Dim fXX(nNumVars) As Double

    For I = 0 To nNumVars - 1
      fXX(I) = fX(I) + fLambda * fDeltaX(I)
    Next I

    MyFxEx = m_MyFx(nNumVars, fXX, fParam)
  End Function

  Protected Function FirstDeriv(ByVal nNumVars As Integer, _
               ByRef fX() As Double, ByRef fParam() As Double, _
               ByVal nIdxI As Integer) As Double
    Dim fXt, h, Fp, Fm As Double

    fXt = fX(nIdxI)
    h = 0.01 * (1 + Math.Abs(fXt))
    fX(nIdxI) = fXt + h
    Fp = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt - h
    Fm = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt
    FirstDeriv = (Fp - Fm) / 2 / h
  End Function

  Protected Sub GetFirstDerives(ByVal nNumVars As Integer, _
                  ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fFirstDerivX() As Double)

    Dim I As Integer

    For I = 0 To nNumVars - 1
      fFirstDerivX(I) = FirstDeriv(nNumVars, fX, fParam, I)
    Next I
  End Sub

  Protected Function LinSearch_DirectSearch(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
            ByRef fLambda As Double, ByRef fDeltaX() As Double, ByVal fInitStep As Double, ByVal fMinStep As Double) As Boolean
    Dim F1, F2 As Double

    F1 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda)

    Do
      F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda + fInitStep)
      If F2 < F1 Then
        F1 = F2
        fLambda += fInitStep
      Else
        F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda - fInitStep)
        If F2 < F1 Then
          F1 = F2
          fLambda -= fInitStep
        Else
          ' reduce search step size
          fInitStep /= 10
        End If
      End If
    Loop Until fInitStep < fMinStep

    Return True

  End Function

  Public Function CalcOptim(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
                    ByRef fToler() As Double, ByVal fEpsFx As Double, ByVal fEpsGrad As Double, ByVal nMaxIter As Integer, _
                    ByRef nIter As Integer, ByRef sErrorMsg As String, _
                    ByVal MyFx As MyFxDelegate) As Double

    Dim I As Integer
    Dim fNorm As Double, fLambda As Double
    Dim bStop As Boolean
    Dim T1, T2 As Double
    ' Declare local arrays
    Dim Index(nNumVars) As Integer
    Dim fGrad1(nNumVars) As Double
    Dim fGrad2(nNumVars) As Double
    Dim g(nNumVars) As Double
    Dim d(nNumVars) As Double
    Dim S(nNumVars) As Double
    Dim Bmat(nNumVars, nNumVars) As Double
    Dim Mmat(nNumVars, nNumVars) As Double
    Dim Nmat(nNumVars, nNumVars) As Double
    Dim MM1(nNumVars, nNumVars) As Double
    Dim MM2(nNumVars, nNumVars) As Double
    Dim MM3(nNumVars, nNumVars) As Double
    Dim MM4(nNumVars, nNumVars) As Double
    Dim MM5(nNumVars, nNumVars) As Double
    Dim MM6(nNumVars, nNumVars) As Double
    Dim MM7(nNumVars, nNumVars) As Double
    Dim MM8(nNumVars, nNumVars) As Double
    Dim MM9(nNumVars, nNumVars) As Double
    Dim MM10(nNumVars, nNumVars) As Double
    Dim MM11(nNumVars, nNumVars) As Double


    ' set error handler
    On Error GoTo HandleErr

    m_MyFx = MyFx
    ' set matrix B as an indentity
    MatrixLibVb.MatIdentity(Bmat)

    nIter = 1
    ' calculate initial gradient
    GetFirstDerives(nNumVars, fX, fParam, fGrad1)

    ' start main loop
    Do

      nIter += 1
      If nIter > nMaxIter Then
        sErrorMsg = "Iter limits"
        Exit Do
      End If

      MatrixLibVb.MatMultVect(Bmat, fGrad1, S)
      MatrixLibVb.VectMultValInPlace(S, -1)

      ' test if gradient is shallow enough
      MatrixLibVb.NormalizeVect(S)

      fLambda = 1
      LinSearch_DirectSearch(nNumVars, fX, fParam, fLambda, S, 0.1, 0.00001)
      ' calculate optimum fX()
      For I = 0 To nNumVars - 1
        d(I) = fLambda * S(I)
        fX(I) += d(I)
      Next I

      ' get new gradient
      GetFirstDerives(nNumVars, fX, fParam, fGrad2)
      For I = 0 To nNumVars - 1
        g(I) = fGrad2(I) - fGrad1(I)
        fGrad1(I) = fGrad2(I)
      Next I

      ' test for convergence
      bStop = True
      For I = 0 To nNumVars - 1
        If Math.Abs(d(I)) > fToler(I) Then
          bStop = False
          Exit For
        End If
      Next I

      If bStop Then
        sErrorMsg = "Lambda * S()"
        Exit Do
      End If

      ' exit if gradient is low
      If MatrixLibVb.VectNorm(g) < fEpsGrad Then
        sErrorMsg = "G Norm"
        Exit Do
      End If

      '-------------------------------------------------
      ' Start elaborate process to upgare matrix B
      '
      MatrixLibVb.VectToColumnMat(g, MM1)        ' MM1 = g as column matrix
      MatrixLibVb.VectToColumnMat(d, MM2)        ' MM2 = d as column matrix
      MatrixLibVb.VectToRowMat(g, MM3)           ' MM3 = g^T
      MatrixLibVb.VectToRowMat(d, MM4)           ' MM4 = d^T
      MatrixLibVb.MatMultMat(MM2, MM4, MM5)      ' MM5 = d d^T
      MatrixLibVb.MatMultMat(MM4, MM1, MM6)      ' MM6 = d^T g
      MatrixLibVb.MatMultMat(MM2, MM3, MM7)      ' MM7 = d g^T
      MatrixLibVb.MatMultMat(MM1, MM4, MM8)      ' MM8 = g d^T
      MatrixLibVb.MatMultMat(MM3, Bmat, MM9)
      MatrixLibVb.MatMultMatInPlace(MM9, MM1)    ' MM9 = g^T [B] g
      MatrixLibVb.MatMultMat(MM7, Bmat, MM10)    ' MM10 = d g^T [B]
      MatrixLibVb.MatMultMat(Bmat, MM8, MM11)    ' MM11 = [B] g d^T

      T1 = MM6(0, 0) ' d^T g
      T2 = (1 + MM9(0, 0) / T1) / T1
      MatrixLibVb.MatMultValInPlace(MM5, T2)     '
      MatrixLibVb.MatMultValInPlace(MM10, 1 / T1)
      MatrixLibVb.MatMultValInPlace(MM11, 1 / T2)
      MatrixLibVb.MatAddMatInPlace(Bmat, MM5)
      MatrixLibVb.MatSubMatInPlace(Bmat, MM10)
      MatrixLibVb.MatSubMatInPlace(Bmat, MM11)

    Loop

ExitProc:
    Return MyFx(nNumVars, fX, fParam)

HandleErr:
    sErrorMsg = "Error: " & Err.Description
    Resume ExitProc
  End Function

End Class

